;; Generated package description from stream.el  -*- no-byte-compile: t -*-
(define-package "stream" "2.3.0.0.20230908.74447" "Implementation of streams" '((emacs "25")) :commit "5c0ffd088fbc344aedb11cc76bd7c0b28f3bc43b" :url "https://elpa.gnu.org/packages/stream.html" :authors '(("Nicolas Petton" . "nicolas@petton.fr")) :maintainer '(nil . "nicolas@petton.fr") :keywords '("stream" "laziness" "sequences"))
